-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("achievementTracker", "AceEvent-3.0", "AceTimer-3.0", "AceBucket-3.0", "AceHook-3.0")

-----------------------
--- Library loading ---
-----------------------
local LSM = LibStub("LibSharedMedia-3.0")
local QTIP = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local CreateFont, CreateFrame, SetFont, ShowUIPanel, GetTime = CreateFont, CreateFrame, SetFont, ShowUIPanel, GetTime
local unpack, format, pairs, ipairs, sub, band, wipe, floor = unpack, string.format, pairs, ipairs, string.sub, bit.band, wipe, math.floor
local tonumber, time, SecondsToTime, ItemRefTooltip = tonumber, time, SecondsToTime, ItemRefTooltip
local GetAchievementInfo, GetAchievementLink, ChatEdit_InsertLink = GetAchievementInfo, GetAchievementLink, ChatEdit_InsertLink
local GetAchievementCategory, GetCategoryInfo, GetCategoryNumAchievements = GetAchievementCategory, GetCategoryInfo, GetCategoryNumAchievements
local GetAchievementNumCriteria, GetAchievementCriteriaInfo = GetAchievementNumCriteria, GetAchievementCriteriaInfo
local UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown = UIParent, IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown

-----------------
--- Variables ---
-----------------
local _
module.numTrackedAchievement = 0
module.CELL_MARGIN_H = 6
module.trackerAnchor = "ANCHOR_RIGHT"
module.cATrFontTQ2 = CreateFont("cATrFontTQ2")
module.scATrFontTQ2 = CreateFont("scATrFontTQ2")
module.aATrFontTQ2 = CreateFont("aATrFontTQ2")
module.acATrFontTQ2 = CreateFont("acATrFontTQ2")
module.sbATrFontTQ2 = CreateFont("sbATrFontTQ2")
module.fixATrFontTQ2 = CreateFont("fixATrFontTQ2")
module.plus = "Interface\\Buttons\\UI-PlusButton-Up"
module.minus = "Interface\\Buttons\\UI-MinusButton-Up"

module.category, module.achievements, module.visibleAchievements, module.visibleCriteria = {}, {}, {}, {}
module.anchorSetPoint = {
	[1] = {point = "TOPLEFT", relPoint = "BOTTOMLEFT", x = 0, y = 24},
	[2] = {point = "TOP", relPoint = "BOTTOM", x = 0, y = 24},
	[3] = {point = "TOPRIGHT", relPoint = "BOTTOMRIGHT", x = 0, y = 24},
	[4] = {point = "BOTTOMLEFT", relPoint = "TOPLEFT", x = 0, y = -24},
	[5] = {point = "BOTTOM", relPoint = "TOP", x = 0, y = -24},
	[6] = {point = "BOTTOMRIGHT", relPoint = "TOPRIGHT", x = 0, y = -24}}
module.achMobs = {1206, 2557, 2556, 1312, 2257}
module.achItems = {1244, 1956, 1225, 1517}
module.achTrade = {1779, 1800}
module.achievementsTradeSkills = {}
module.substitute = {["Écureuil"] = "Ecureuil", ["Écureuil des Grisonnes"] = "Ecureuil des Grisonnes"}

-----------------------
--- Tools functions ---
-----------------------
local function normalizeWidth(min, max)
	if max < min then return max else return min end
end

local function normalizeColorComponent(component)
	if component > 1 then return 1 elseif component < 0 then return 0 else return component end
end

local function getColorFromCompletionRGB(percent)
	if percent <= 0 then
		return core.db.profile.quests.rj0, core.db.profile.quests.gj0, core.db.profile.quests.bj0
	elseif percent >= 1 then
		return core.db.profile.quests.rj3, core.db.profile.quests.gj3, core.db.profile.quests.bj3
	elseif percent <= 0.5 then
		return normalizeColorComponent(core.db.profile.quests.ri1*percent + core.db.profile.quests.rj1), normalizeColorComponent(core.db.profile.quests.gi1*percent + core.db.profile.quests.gj1), normalizeColorComponent(core.db.profile.quests.bi1*percent + core.db.profile.quests.bj1)
	else
		return normalizeColorComponent(core.db.profile.quests.ri2*percent + core.db.profile.quests.rj2), normalizeColorComponent(core.db.profile.quests.gi2*percent + core.db.profile.quests.gj2), normalizeColorComponent(core.db.profile.quests.bi2*percent + core.db.profile.quests.bj2)
	end
end

local function getColorFromCompletionHEX(percent)
	if percent <= 0 then
		return format("%02x%02x%02x", 255*core.db.profile.quests.rj0, 255*core.db.profile.quests.gj0, 255*core.db.profile.quests.bj0)
	elseif percent >= 1 then
		return format("%02x%02x%02x", 255*core.db.profile.quests.rj3, 255*core.db.profile.quests.gj3, 255*core.db.profile.quests.bj3)
	elseif percent <= 0.5 then
		return format("%02x%02x%02x", 255*normalizeColorComponent(core.db.profile.quests.ri1*percent + core.db.profile.quests.rj1), 255*normalizeColorComponent(core.db.profile.quests.gi1*percent + core.db.profile.quests.gj1), 255*normalizeColorComponent(core.db.profile.quests.bi1*percent + core.db.profile.quests.bj1))
	else
		return format("%02x%02x%02x", 255*normalizeColorComponent(core.db.profile.quests.ri2*percent + core.db.profile.quests.rj2), 255*normalizeColorComponent(core.db.profile.quests.gi2*percent + core.db.profile.quests.gj2), 255*normalizeColorComponent(core.db.profile.quests.bi2*percent + core.db.profile.quests.bj2))
	end
end

local function setUnitTooltip()
	for _, aid in pairs(module.achMobs) do
		local _, name, _, completed, _, _, _, _, _, icon = GetAchievementInfo(aid)
		if not completed then
			core.achievementName[aid] = name
			core.achievementIcon[aid] = icon
			local numCriteria = GetAchievementNumCriteria(aid)
			for cid = 1, numCriteria do
				local cString, _, completed = GetAchievementCriteriaInfo(aid, cid)
				if module.substitute[cString] then cString = module.substitute[cString] end
				if not core.achievementsMobs[cString] then
					core.achievementsMobs[cString] = core.new()
				end
				core.achievementsMobs[cString][#(core.achievementsMobs[cString])+1] = aid.."~"..cid
			end
		end
	end
end

local function setItemTooltip()
	for _, aid in pairs(module.achItems) do
		local _, name, _, completed, _, _, _, _, _, icon = GetAchievementInfo(aid)
		if not completed then
			core.achievementName[aid] = name
			core.achievementIcon[aid] = icon
			local numCriteria = GetAchievementNumCriteria(aid)
			for cid = 1, numCriteria do
				local cString, _, completed = GetAchievementCriteriaInfo(aid, cid)
				if not core.achievementsItems[cString] then
					core.achievementsItems[cString] = core.new()
				end
				core.achievementsItems[cString][#(core.achievementsItems[cString])+1] = aid.."~"..cid
			end
		end
	end
end

local function setTSUI()
	for _, aid in pairs(module.achTrade) do
		local _, name, _, completed, _, _, _, _, _, icon = GetAchievementInfo(aid)
		if not completed then
			core.achievementName[aid] = name
			core.achievementIcon[aid] = icon
			local numCriteria = GetAchievementNumCriteria(aid)
			for cid = 1, numCriteria do
				local cString, _, completed = GetAchievementCriteriaInfo(aid, cid)
				if not completed then
					module.achievementsTradeSkills[cString] = aid
				end
			end
			module.hookTradeSkills = true
		end
	end
	if module.hookTradeSkills then
		module.cooking = GetCategoryInfo(170)
		if IsAddOnLoaded("Blizzard_TradeSkillUI") then
			module:SecureHook("TradeSkillFrame_Update")
		elseif IsAddOnLoaded("MrTrader_SkillWindow") then
			module:SecureHook(MRTSkillWindow, "UpdateSkillList")
		elseif IsAddOnLoaded("Skillet") then
			module:SecureHook(Skillet, "internal_UpdateTradeSkillWindow")
		elseif IsAddOnLoaded("AdvancedTradeSkillWindow") then
			module:SecureHook("ATSWFrame_Update")
		else
			module:RegisterEvent("ADDON_LOADED")
		end
	end
end

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		timedAchievements = true,
		point = "CENTER",
		x = 0,
		y = 0,
		showInQuestTracker = false,
		positionInQuestTracker = 1,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		hideCompletedCriteria = true,
		showByCategory = true,
		categoryFont = LSM.DefaultMedia.font,
		categoryFontSize = 12,
		categoryFontOutline = "",
		categoryCollapseButtonSize =13,
		subCategoryFont = LSM.DefaultMedia.font,
		subCategoryFontSize = 12,
		subCategoryFontOutline = "",
		subCategoryCollapseButtonSize = 13,
		achievementFont = LSM.DefaultMedia.font,
		achievementFontSize = 10,
		achievementFontOutline = "",
		displayAchievementImage = true,
		achievementImageSize = 13,
		achievementCriteriaFont = LSM.DefaultMedia.font,
		achievementCriteriaFontSize = 8,
		achievementCriteriaFontOutline = "",
		statusBarFont = LSM.DefaultMedia.font,
		statusBarFontSize = 8,
		statusBarFontOutline = "",
		numBooleanCriteriaColumns = 2,
		maxWidth = 250,
		minWidth = 250,
		statusMaxWidth = 70,
		statusMinWidth = 70,
		hasMaxHeight = true,
		maxHeight = 400,
		subCategoryCellPadding = 5,
		achievementCellPadding = 5,
		criteriaCellPadding = 10,
		rightCellPadding = 0,
		scale = 1,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
		colorizeACName = true,
		defaultACNameColor = "ffffff",
		colorizeACCounter = true,
		defaultACCounterColor = "ffffff",
		colorizeASCName = true,
		defaultASCNameColor = "ffffff",
		colorizeASCCounter = true,
		defaultASCCounterColor = "ffffff",
		colorizeAname = true,
		defaultANameColor = "ffffff",
		colorizeATimer = true,
		defaultATimerColor = "ffffff",
		colorizeACrName = true,
		defaultACrNameColor = "ffffff",
		colorizeACrStatus = true,
		defaultACrStatusColor = "ffffff",
		displayAsStatusBar = true,
		statusBarTexture = LSM.DefaultMedia.statusbar,
		colorizeStatusBar = true,
		defaultStatusBarColor = {0, 0.6, 0},
		colorizeStatusBarText = false,
		defaultStatusBarTextColor = "ffffff",
		defaultDescriptionColor = "ffffff",
		showSBDescription = true,
		statusBarHeight = 12,
		leftStatusBarPadding = 40,
		rightStatusBarPadding = 40,
		disableInteraction = false,
		overTooltip = true,
		mouseRightShift = "chatLink",
		mouseRightControl = "nothing",
		mouseRightAlt = "nothing",
		mouseRight = "achievementTooltip",
		mouseLeftShift = "removeTrackedAchievement",
		mouseLeftControl = "nothing",
		mouseLeftAlt = "nothing",
		mouseLeft = "displayAchievement",
		mouseMiddleShift = "nothing",
		mouseMiddleControl = "nothing",
		mouseMiddleAlt = "nothing",
		mouseMiddle = "nothing",
		unitTooltip = true,
		itemTooltip = true,
		tradeSkillUI = true,
	},
}

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("achievementTracker", defaults, "Default")
	self.db.char.achievementTracked = self.db.char.achievementTracked or {}
	self.db.char.categoryDisplay = self.db.char.categoryDisplay or {}
	self:SetEnabledState(core:getModuleEnabled("achievementTracker"))
	core.moduleState["achievementTracker"] = core:getModuleEnabled("achievementTracker")
end

function module:OnEnable()
	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "updateUsedMedia")
	self.cATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.categoryFont), module.db.profile.categoryFontSize, module.db.profile.categoryFontOutline)
	self.scATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.subCategoryFont), module.db.profile.subCategoryFontSize, self.db.profile.subCategoryFontOutline)
	self.aATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.achievementFont), self.db.profile.achievementFontSize, self.db.profile.achievementFontOutline)
	self.acATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.achievementCriteriaFont), self.db.profile.achievementCriteriaFontSize, self.db.profile.achievementCriteriaFontOutline)
	self.sbATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.statusBarFont), self.db.profile.statusBarFontSize, self.db.profile.statusBarFontOutline)
	self.fixATrFontTQ2:SetFont(LSM:Fetch("font", LSM.DefaultMedia.font), 1)
	self.anchor = CreateFrame("Button", nil, UIParent)
	self.anchor:SetHeight(24)
	self.anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	self.anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
	self.anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
	self.anchor.text = self.anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
	self.anchor.text:SetPoint("CENTER")
	self.anchor.text:SetText(core.LC["Achievement Tracker"])
	self.anchor:SetWidth(self.anchor.text:GetStringWidth() + 8)
	self.anchor:SetMovable(true)
	self.anchor:RegisterForDrag("LeftButton")
	self.anchor:SetClampedToScreen(true)
	self.anchor:SetPoint(self.db.profile.point, self.db.profile.x, self.db.profile.y)
	self.anchor:SetScript("OnDragStart", function(self) self:StartMoving() end)
	self.anchor:SetScript("OnDragStop", function(self)
		self:StopMovingOrSizing()
		module.db.profile.point, module.db.profile.x, module.db.profile.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
		module.trackerAnchor = module.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
		module:updateAchievementTracker()
	end)
	if not self.db.profile.showAnchor then self.anchor:Hide() else self.anchor:Show() end
	self.trackerAnchor = self.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if not self.db.profile.showInQuestTracker or not core.moduleState["questsTracker"] then
		self.disableInteraction = self.db.profile.disableInteraction
		self.maxWidth = self.db.profile.maxWidth
		self.statusMaxWidth = self.db.profile.statusMaxWidth
		self.columnMaxWidth = self.maxWidth / self.db.profile.numBooleanCriteriaColumns
		self.totalMaxWidth = self.maxWidth + self.statusMaxWidth + self.CELL_MARGIN_H * self.db.profile.numBooleanCriteriaColumns
		self.minWidth = normalizeWidth(self.db.profile.minWidth, self.maxWidth)
		self.statusMinWidth = normalizeWidth(self.db.profile.statusMinWidth, self.statusMaxWidth)
		self.columnMinWidth = normalizeWidth(self.minWidth / self.db.profile.numBooleanCriteriaColumns, self.columnMaxWidth)
		self.totalMinWidth = normalizeWidth(self.minWidth + self.statusMinWidth + self.CELL_MARGIN_H * self.db.profile.numBooleanCriteriaColumns, self.totalMaxWidth)
	else
		core.cbh:Fire("updateAchievementsTrackerParameters")
	end
	if self.db.profile.displayAchievementImage then
		self.achievementImageSize = self.db.profile.achievementImageSize
		self.achievementCellPadding = self.db.profile.achievementCellPadding + self.achievementImageSize
	else
		self.achievementImageSize = 0
		self.achievementCellPadding = self.db.profile.achievementCellPadding
	end
	self.categoryCellPadding = self.db.profile.categoryCollapseButtonSize
	self.subCategoryCellPadding = self.db.profile.subCategoryCellPadding + self.db.profile.subCategoryCollapseButtonSize
	self.achievementIconPadding = self.db.profile.achievementCellPadding
	self.criteriaCellPadding = self.db.profile.criteriaCellPadding
	if self.db.profile.showByCategory then
		self.subAchievementCellPadding = self.achievementCellPadding + self.subCategoryCellPadding
		self.subAchievementIconPadding = self.db.profile.achievementCellPadding + self.subCategoryCellPadding
		self.subCriteriaCellPadding = self.criteriaCellPadding + self.subCategoryCellPadding
	else
		self.subAchievementCellPadding = self.achievementCellPadding
		self.subAchievementIconPadding = self.achievementIconPadding
		self.subCriteriaCellPadding = self.criteriaCellPadding
	end
	self:SecureHook("AddTrackedAchievement")
	self:SecureHook("RemoveTrackedAchievement")
	self:RawHook("GetNumTrackedAchievements", true)
	self:RawHook("IsTrackedAchievement", true)
	self:RegisterBucketEvent("CRITERIA_UPDATE", 1)
	self:RegisterEvent("ACHIEVEMENT_EARNED")
	core.RegisterCallback(self, "updateAchievementTracker")
	if self.db.profile.timedAchievements then
		self:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
	end
	if self.db.profile.hideTrackerInCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "outOfCombat")
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "inCombat")
	end
	for aid in pairs(self.db.char.achievementTracked) do
		local _, _, _, completed = GetAchievementInfo(aid)
		if not completed then
			AddTrackedAchievement(aid, true)
		else
			self.db.char.achievementTracked[aid] = nil
		end
	end
	if AchievementFrame then
		AchievementFrameAchievements_ForceUpdate()
	end
	if self.db.profile.unitTooltip then
		setUnitTooltip()
	end
	if self.db.profile.itemTooltip then
		setItemTooltip()
	end
	if self.db.profile.tradeSkillUI then
		setTSUI()
	end
	self:updateAchievementTracker()
end

function module:OnDisable()
	self.anchor:Hide()
	QTIP:Release(self.tracker)
	self.tracker = nil
	self:Refresh()
end

function module:Refresh()
	if not core.moduleState["achievementTracker"] then return end
	self:UnregisterAllEvents()
	self.cATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.categoryFont), self.db.profile.categoryFontSize, self.db.profile.categoryFontOutline)
	self.scATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.subCategoryFont), self.db.profile.subCategoryFontSize, self.db.profile.subCategoryFontOutline)
	self.aATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.achievementFont), self.db.profile.achievementFontSize, self.db.profile.achievementFontOutline)
	self.acATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.achievementCriteriaFont), self.db.profile.achievementCriteriaFontSize, self.db.profile.achievementCriteriaFontOutline)
	self.sbATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.statusBarFont), self.db.profile.statusBarFontSize, self.db.profile.statusBarFontOutline)
	self.fixATrFontTQ2:SetFont(LSM:Fetch("font", LSM.DefaultMedia.font), 1)
	self.anchor:SetPoint(self.db.profile.point, self.db.profile.x, self.db.profile.y)
	if not self.db.profile.showAnchor then self.anchor:Hide() else self.anchor:Show() end
	self.trackerAnchor = self.anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if not self.db.profile.showInQuestTracker or not core.moduleState["questsTracker"] then
		self.disableInteraction = self.db.profile.disableInteraction
		self.maxWidth = self.db.profile.maxWidth
		self.statusMaxWidth = self.db.profile.statusMaxWidth
		self.columnMaxWidth = self.maxWidth / self.db.profile.numBooleanCriteriaColumns
		self.totalMaxWidth = self.maxWidth + self.statusMaxWidth + self.CELL_MARGIN_H * self.db.profile.numBooleanCriteriaColumns
		self.minWidth = normalizeWidth(self.db.profile.minWidth, self.maxWidth)
		self.statusMinWidth = normalizeWidth(self.db.profile.statusMinWidth, self.statusMaxWidth)
		self.columnMinWidth = normalizeWidth(self.minWidth / self.db.profile.numBooleanCriteriaColumns, self.columnMaxWidth)
		self.totalMinWidth = normalizeWidth(self.minWidth + self.statusMinWidth + self.CELL_MARGIN_H * self.db.profile.numBooleanCriteriaColumns, self.totalMaxWidth)
	else
		core.cbh:Fire("updateAchievementsTrackerParameters")
	end
	if self.db.profile.displayAchievementImage then
		self.achievementImageSize = self.db.profile.achievementImageSize
		self.achievementCellPadding = self.db.profile.achievementCellPadding + self.achievementImageSize
	else
		self.achievementImageSize = 0
		self.achievementCellPadding = self.db.profile.achievementCellPadding
	end
	self.categoryCellPadding = self.db.profile.categoryCollapseButtonSize
	self.subCategoryCellPadding = self.db.profile.subCategoryCellPadding + self.db.profile.subCategoryCollapseButtonSize
	self.achievementIconPadding = self.db.profile.achievementCellPadding
	self.criteriaCellPadding = self.db.profile.criteriaCellPadding
	if self.db.profile.showByCategory then
		self.subAchievementCellPadding = self.achievementCellPadding + self.subCategoryCellPadding
		self.subAchievementIconPadding = self.db.profile.achievementCellPadding + self.subCategoryCellPadding
		self.subCriteriaCellPadding = self.criteriaCellPadding + self.subCategoryCellPadding
	else
		self.subAchievementCellPadding = self.achievementCellPadding
		self.subAchievementIconPadding = self.achievementIconPadding
		self.subCriteriaCellPadding = self.criteriaCellPadding
	end
	if self.db.profile.timedAchievements then
		self:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
	end
	if self.db.profile.hideTrackerInCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "outOfCombat")
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "inCombat")
	end
	self:updateAchievementTracker()
end

-----------------
--- Callbacks ---
-----------------
function module:inCombat(event)
	self.hideInCombat = true
	if self.tracker then self.tracker:Hide() end
end

function module:outOfCombat(event)
	self.hideInCombat = nil
	self:updateAchievementTracker()
end

function module:updateUsedMedia(event, mediaType, key)
	if not core.moduleState["achievementTracker"] then return end
	if mediaType == "font" then
		local updateTracker = nil
		if key == self.db.profile.categoryFont then
			updateTracker = true
			self.cATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.categoryFont), self.db.profile.categoryFontSize, self.db.profile.categoryFontOutline)
		end
		if key == self.db.profile.subCategoryFont then
			updateTracker = true
			self.scATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.subCategoryFont), self.db.profile.subCategoryFontSize, self.db.profile.subCategoryFontOutline)
		end
		if key == self.db.profile.achievementFont then
			updateTracker = true
			self.aATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.achievementFont), self.db.profile.achievementFontSize, self.db.profile.achievementFontOutline)
		end
		if key == self.db.profile.achievementCriteriaFont then
			updateTracker = true
			self.acATrFontTQ2:SetFont(LSM:Fetch("font", self.db.profile.achievementCriteriaFont), self.db.profile.achievementCriteriaFontSize, self.db.profile.achievementCriteriaFontOutline)
		end
		if updateTracker then
			self:updateAchievementTracker()
		end
	end
end

--------------
--- Events ---
--------------
function module:ADDON_LOADED(event, addon)
	if addon == "Blizzard_TradeSkillUI" then
		self:SecureHook("TradeSkillFrame_Update")
		self:UnregisterEvent("ADDON_LOADED")
	elseif addon == "MrTrader_SkillWindow" then
		self:SecureHook(MRTSkillWindow, "UpdateSkillList")
		self:UnregisterEvent("ADDON_LOADED")
	elseif addon == "Skillet" then
		self:SecureHook(Skillet, "internal_UpdateTradeSkillWindow")
		self:UnregisterEvent("ADDON_LOADED")
	elseif addon == "AdvancedTradeSkillWindow" then
		self:SecureHook("ATSWFrame_Update")
		self:UnregisterEvent("ADDON_LOADED")
	end
end

function module:ACHIEVEMENT_EARNED(event, aid)
	if self.achievements[aid] then
		RemoveTrackedAchievement(aid)
	end
	local subCategory = GetAchievementCategory(aid)
	local subCategoryName, category = GetCategoryInfo(subCategory)
	if category and category > 0 then
		if self.category[category] then
			self.category[category].numTA, self.category[category].numTCA  = GetCategoryNumAchievements(category)
			self.category[category].color = getColorFromCompletionHEX(self.category[category].numTCA / self.category[category].numTA)
		end
		if self.category[subCategory] then
			self.category[subCategory].numTA, self.category[subCategory].numTCA  = GetCategoryNumAchievements(subCategory)
			self.category[subCategory].color = getColorFromCompletionHEX(self.category[subCategory].numTCA / self.category[subCategory].numTA)
		end
	elseif self.category[subCategory] then
		self.category[subCategory].numTA, self.category[subCategory].numTCA  = GetCategoryNumAchievements(subCategory)
		self.category[subCategory].color = getColorFromCompletionHEX(self.category[subCategory].numTCA / self.category[subCategory].numTA)
	end
end

function module:CRITERIA_UPDATE()
	local updateTracker
	for aid in pairs(self.achievements) do
		for cid = 1, self.achievements[aid].numCriteria do
			local _, _, completed, quantity, reqQuantity = GetAchievementCriteriaInfo(aid, cid)
			completed = completed or 0
			if quantity ~= self.achievements[aid].criteria[cid].quantity or completed ~= self.achievements[aid].criteria[cid].completed then
				self.achievements[aid].criteria[cid].completed = completed
				self.achievements[aid].criteria[cid].quantity = quantity
				self.achievements[aid].criteria[cid].color = getColorFromCompletionHEX(quantity / self.achievements[aid].criteria[cid].reqQuantity)
				self.achievements[aid].criteria[cid].rColor, self.achievements[aid].criteria[cid].gColor, self.achievements[aid].criteria[cid].bColor = getColorFromCompletionRGB(quantity / self.achievements[aid].criteria[cid].reqQuantity)
				if self.visibleCriteria[self.achievements[aid].criteria[cid].id] then
					self:updateAchievementTracker()
				end
			end
		end
	end
	for _, aid in pairs(self.achTrade) do
		local numCriteria = GetAchievementNumCriteria(aid)
		for cid = 1, numCriteria do
			local cString, _, completed = GetAchievementCriteriaInfo(aid, cid)
			if completed and self.achievementsTradeSkills[cString] then
				self.achievementsTradeSkills[cString] = nil
			end
		end
	end
	if updateTracker then
		self:updateAchievementTracker()
	end
end

function module:TRACKED_ACHIEVEMENT_UPDATE(event, aid, cid, elapsed, duration)
	if not aid or not duration or not elapsed then return end
	local startTime = GetTime()
	if duration > elapsed then
		if not self.achievements[aid] then
			AddTrackedAchievement(aid)
		end
		self.achievements[aid].startTime = startTime - elapsed
		self.achievements[aid].duration = duration
	elseif self.achievements[aid] and self.achievements[aid].startTime then
		self.achievements[aid].startTime = nil
		self.achievements[aid].duration = nil
	end
	if self.visibleAchievements[aid] then
		self:updateAchievementTracker()
	end
end

-------------
--- Hooks ---
-------------
local function tradeSkillOnEnter(self)
	if self.aid and core.achievementIcon[self.aid] and core.achievementName[self.aid] then
		GameTooltip:SetOwner(self, "ANCHOR_NONE")
		GameTooltip:SetPoint("TOPLEFT", self, "BOTTOMLEFT")
		GameTooltip:AddLine("|T"..core.achievementIcon[self.aid]..":0|t "..core.achievementName[self.aid], 1, 0, 0)
		GameTooltip:Show()
	end
end

local function tradeSkillOnLeave(self)
	GameTooltip:Hide()
end

function module:TradeSkillFrame_Update()
	local tradeSkillName = GetTradeSkillLine()
	if tradeSkillName == self.cooking then
		local offset = FauxScrollFrame_GetOffset(TradeSkillListScrollFrame)
		for i = 1, TRADE_SKILLS_DISPLAYED do
			local skillName, skillType = GetTradeSkillInfo(i + offset)
			if skillType ~= "header" then
				local button = _G["TradeSkillSkill"..i]
				button.aid = self.achievementsTradeSkills[skillName]
				if button.aid then
					button:SetNormalTexture(core.achievementIcon[self.achievementsTradeSkills[skillName]])
					_G["TradeSkillSkill"..i.."Highlight"]:SetTexture(core.achievementIcon[self.achievementsTradeSkills[skillName]])
					button:SetScript("OnEnter", tradeSkillOnEnter)
					button:SetScript("OnLeave", tradeSkillOnLeave)
				else
					button:SetNormalTexture("")
					_G["TradeSkillSkill"..i.."Highlight"]:SetTexture("")
				end
			end
		end
	end
end

local function mrtButtonOnEnter(self)
	if self.aid and core.achievementIcon[self.aid] and core.achievementName[self.aid] then
		GameTooltip:AddLine("|T"..core.achievementIcon[self.aid]..":0|t "..core.achievementName[self.aid], 1, 0, 0)
		GameTooltip:Show()
	end
end

function module:UpdateSkillList()
	local tradeSkillName = GetTradeSkillLine()
	if tradeSkillName == self.cooking then
		local offset = FauxScrollFrame_GetOffset(MRTSkillListScrollFrame)
		for i = 1, 11 do
			local skillName, skillType = GetTradeSkillInfo(i + offset)
			if skillType ~= "header" then
				local button = _G["MRTSkillButton"..i]
				button.aid = self.achievementsTradeSkills[skillName]
				if button.aid then
					button:SetNormalTexture(core.achievementIcon[self.achievementsTradeSkills[skillName]])
					_G["MRTSkillButton"..i.."Highlight"]:SetTexture(core.achievementIcon[self.achievementsTradeSkills[skillName]])
					if not button.hooked then
						button:HookScript("OnEnter", mrtButtonOnEnter)
						button.hooked = true
					end
				else
					button:SetNormalTexture("")
					_G["MRTSkillButton"..i.."Highlight"]:SetTexture("")
				end
			end
		end
	end
end

local function skilletButtonOnEnter(self)
	if self.aid and core.achievementIcon[self.aid] and core.achievementName[self.aid] then
		SkilletTradeskillTooltip:AddLine("|T"..core.achievementIcon[self.aid]..":0|t "..core.achievementName[self.aid], 1, 0, 0)
		SkilletTradeskillTooltip:Show()
	end
end

function module:internal_UpdateTradeSkillWindow()
	local tradeSkillName = GetTradeSkillLine()
	if tradeSkillName == self.cooking then
		local offset = FauxScrollFrame_GetOffset(SkilletSkillList)
		for i = 1, floor(SkilletSkillList:GetHeight() / SKILLET_TRADE_SKILL_HEIGHT) do
			local skillName, skillType = GetTradeSkillInfo(i + offset)
			if skillType ~= "header" then
				local button = _G["SkilletScrollButton"..i]
				button.aid = self.achievementsTradeSkills[skillName]
				if button.aid then
					button:SetNormalTexture(core.achievementIcon[self.achievementsTradeSkills[skillName]])
					_G["SkilletScrollButton"..i.."Highlight"]:SetTexture(core.achievementIcon[self.achievementsTradeSkills[skillName]])
					if not button.hooked then
						button:HookScript("OnEnter", skilletButtonOnEnter)
						button.hooked = true
					end
				else
					button:SetNormalTexture("")
					_G["SkilletScrollButton"..i.."Highlight"]:SetTexture("")
				end
				_G["SkilletScrollButton"..i.."Level"]:SetWidth(20)
			end
		end
	end
end

local function atswButtonOnEnter(self)
	if self.aid and core.achievementIcon[self.aid] and core.achievementName[self.aid] then
		ATSWTradeskillTooltip:AddLine("|T"..core.achievementIcon[self.aid]..":0|t "..core.achievementName[self.aid], 1, 0, 0)
		ATSWTradeskillTooltip:Show()
	end
end

function module:ATSWFrame_Update()
	if not ATSWFrame:IsVisible() then  return end
	local tradeSkillName = GetTradeSkillLine()
	if tradeSkillName == self.cooking then
		local offset = FauxScrollFrame_GetOffset(ATSWListScrollFrame)
		for i = 1, ATSW_TRADE_SKILLS_DISPLAYED do
			local skillName, skillType = GetTradeSkillInfo(i + offset)
			if skillType ~= "header" then
				local button = _G["ATSWSkill"..i]
				button.aid = self.achievementsTradeSkills[skillName]
				if button.aid then
					button:SetNormalTexture(core.achievementIcon[self.achievementsTradeSkills[skillName]])
					_G["ATSWSkill"..i.."Highlight"]:SetTexture(core.achievementIcon[self.achievementsTradeSkills[skillName]])
					if not button.hooked then
						button:HookScript("OnEnter", atswButtonOnEnter)
						button.hooked = true
					end
				else
					button:SetNormalTexture("")
					_G["ATSWSkill"..i.."Highlight"]:SetTexture("")
				end
			end
		end
	end
end

function module:AddTrackedAchievement(aid, batch)
	if not batch and self.db.char.achievementTracked[aid] then
		RemoveTrackedAchievement(aid)
	else
		self.db.char.achievementTracked[aid] = true
		self.numTrackedAchievement = self.numTrackedAchievement + 1
		self.achievements[aid] = core.new()
		_, self.achievements[aid].name, _, self.achievements[aid].completed, _, _, _, self.achievements[aid].description, _, self.achievements[aid].image = GetAchievementInfo(aid)
		self.achievements[aid].numCriteria = GetAchievementNumCriteria(aid)
		self.achievements[aid].criteria = core.new()
		for cid = 1, self.achievements[aid].numCriteria do
			local cString, cType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, cid2 = GetAchievementCriteriaInfo(aid, cid)
			if reqQuantity == 0 then reqQuantity = 1 end
			self.achievements[aid].criteria[cid] = core.new()
			self.achievements[aid].criteria[cid].id = cid2
			self.achievements[aid].criteria[cid].name = cString
			self.achievements[aid].criteria[cid].cType = cType
			self.achievements[aid].criteria[cid].completed = completed or 0
			self.achievements[aid].criteria[cid].quantity = quantity
			self.achievements[aid].criteria[cid].reqQuantity = reqQuantity
			self.achievements[aid].criteria[cid].color = getColorFromCompletionHEX(quantity / reqQuantity)
			self.achievements[aid].criteria[cid].rColor, self.achievements[aid].criteria[cid].gColor, self.achievements[aid].criteria[cid].bColor = getColorFromCompletionRGB(quantity / reqQuantity)
			self.achievements[aid].criteria[cid].statusBar = band(flags, ACHIEVEMENT_CRITERIA_PROGRESS_BAR) == ACHIEVEMENT_CRITERIA_PROGRESS_BAR
		end
		local categoryName, subCategoryName
		self.achievements[aid].subCategory = GetAchievementCategory(aid)
		subCategoryName, self.achievements[aid].category = GetCategoryInfo(self.achievements[aid].subCategory)
		categoryName = GetCategoryInfo(self.achievements[aid].category)
		if self.achievements[aid].category and self.achievements[aid].category > 0 then
			if not self.category[self.achievements[aid].category] then
				self.category[self.achievements[aid].category] = core.new()
				self.category[self.achievements[aid].category].subCategory = core.new()
				self.category[self.achievements[aid].category].numSubCategory = 0
				self.category[self.achievements[aid].category].achievements = core.new()
				self.category[self.achievements[aid].category].numAchievements = 0
				self.category[self.achievements[aid].category].name = categoryName
				self.category[self.achievements[aid].category].numTA, self.category[self.achievements[aid].category].numTCA  = GetCategoryNumAchievements(self.achievements[aid].category)
				self.category[self.achievements[aid].category].color = getColorFromCompletionHEX(self.category[self.achievements[aid].category].numTCA / self.category[self.achievements[aid].category].numTA)
			end
			if not self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory] then
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory] = core.new()
				self.category[self.achievements[aid].category].numSubCategory = self.category[self.achievements[aid].category].numSubCategory + 1
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements = 0
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].achievements = core.new()
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].name = subCategoryName
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numTA, self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numTCA  = GetCategoryNumAchievements(self.achievements[aid].subCategory)
				self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].color = getColorFromCompletionHEX(self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numTCA / self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numTA)
			end
			self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements = self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements + 1
			self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].achievements[aid] = true
			self.achievements[aid].sc = true
		else
			if not self.category[self.achievements[aid].subCategory] then
				self.category[self.achievements[aid].subCategory] = core.new()
				self.category[self.achievements[aid].subCategory].subCategory = core.new()
				self.category[self.achievements[aid].subCategory].numSubCategory = 0
				self.category[self.achievements[aid].subCategory].achievements = core.new()
				self.category[self.achievements[aid].subCategory].numAchievements = 0
				self.category[self.achievements[aid].subCategory].name = subCategoryName
				self.category[self.achievements[aid].subCategory].numTA, self.category[self.achievements[aid].subCategory].numTCA  = GetCategoryNumAchievements(self.achievements[aid].subCategory)
				self.category[self.achievements[aid].subCategory].color = getColorFromCompletionHEX(self.category[self.achievements[aid].subCategory].numTCA / self.category[self.achievements[aid].subCategory].numTA)
			end
			self.category[self.achievements[aid].subCategory].numAchievements = self.category[self.achievements[aid].subCategory].numAchievements + 1
			self.category[self.achievements[aid].subCategory].achievements[aid] = true
		end
		if not batch then
			if AchievementFrame then
				AchievementFrameAchievements_ForceUpdate()
			end
			self:updateAchievementTracker()
		end
	end
end

function module:RemoveTrackedAchievement(aid)
	if not self.db.char.achievementTracked[aid] or not self.achievements[aid] then return end
	self.numTrackedAchievement = self.numTrackedAchievement - 1
	if self.achievements[aid].startTime then
		self.achievements[aid].startTime = nil
		self.achievements[aid].duration = nil
	end
	self.db.char.achievementTracked[aid] = nil
	if self.achievements[aid].category and self.achievements[aid].category > 0 and self.achievements[aid].subCategory and self.achievements[aid].subCategory > 0 then
		self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].achievements[aid] = nil
		self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements = self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements - 1
		if self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory].numAchievements == 0 then
			self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory] = core.del(self.category[self.achievements[aid].category].subCategory[self.achievements[aid].subCategory])
			self.category[self.achievements[aid].category].numSubCategory = self.category[self.achievements[aid].category].numSubCategory - 1
			if self.category[self.achievements[aid].category].numSubCategory == 0 and self.category[self.achievements[aid].category].numAchievements == 0 then
				self.category[self.achievements[aid].category] = core.del(self.category[self.achievements[aid].category])
			end
		end
	elseif self.achievements[aid].subCategory and self.achievements[aid].subCategory > 0 then
		self.category[self.achievements[aid].subCategory].achievements[aid] = nil
		self.category[self.achievements[aid].subCategory].numAchievements = self.category[self.achievements[aid].subCategory].numAchievements - 1
		if self.category[self.achievements[aid].subCategory].numSubCategory == 0 and self.category[self.achievements[aid].subCategory].numAchievements == 0 then
			self.category[self.achievements[aid].subCategory] = core.del(self.category[self.achievements[aid].subCategory])
		end
	end
	self.achievements[aid] = core.del(self.achievements[aid])
	self:updateAchievementTracker()
	if AchievementFrame then
		AchievementFrameAchievements_ForceUpdate()
	end
end

function module:GetNumTrackedAchievements()
	return self.numTrackedAchievement
end

function module:IsTrackedAchievement(aid)
	return self.achievements[aid] and true or false
end

---------------
--- Tracker ---
---------------
function module:ReleaseTracker()
	QTIP:Release(self.tracker)
	self.tracker = nil
end

local interactivity = {
	chatLink = function(cValue1)
		ChatFrameEditBox:Show()
		ChatEdit_InsertLink(GetAchievementLink(cValue1))
	end,
	achievementTooltip = function(cValue1)
		ItemRefTooltip:ClearLines()
		ShowUIPanel(ItemRefTooltip)
		if not ItemRefTooltip:IsVisible() then
			ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
		end
		local link = GetAchievementLink(cValue1)
		if link then
			ItemRefTooltip:SetHyperlink(link)
		end
	end,
	removeTrackedAchievement = function(cValue1)
		RemoveTrackedAchievement(cValue1)
	end,
	displayAchievement = function(cValue1)
		ToggleAchievementFrame()
		if AchievementFrame:IsShown() then
			AchievementFrame_SelectAchievement(cValue1)
		end
	end,
	nothing = function(cValue1)
		return
	end,
}

local function achievementOnMouseDown(cell, button)
	local cType1, cValue1 = cell.cType1, cell.cValue1
	if cType1 == "category" then
		if module.db.char.categoryDisplay[cValue1] == 1 then
			module.db.char.categoryDisplay[cValue1] = 0
		else
			module.db.char.categoryDisplay[cValue1] = 1
		end
		module:updateAchievementTracker()
	elseif cType1 == "achievement" then
		if button == "RightButton" then
			if IsShiftKeyDown() then
				interactivity[module.db.profile.mouseRightShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[module.db.profile.mouseRightControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[module.db.profile.mouseRightAlt](cValue1)
			else
				interactivity[module.db.profile.mouseRight](cValue1)
			end
		elseif button == "LeftButton" then
			if IsShiftKeyDown() then
				interactivity[module.db.profile.mouseLeftShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[module.db.profile.mouseLeftControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[module.db.profile.mouseLeftAlt](cValue1)
			else
				interactivity[module.db.profile.mouseLeft](cValue1)
			end
		elseif button == "MiddleButton" then
			if IsShiftKeyDown() then
				interactivity[module.db.profile.mouseMiddleShift](cValue1)
			elseif IsControlKeyDown() then
				interactivity[module.db.profile.mouseMiddleControl](cValue1)
			elseif IsAltKeyDown() then
				interactivity[module.db.profile.mouseMiddleAlt](cValue1)
			else
				interactivity[module.db.profile.mouseMiddle](cValue1)
			end
		end
	end
end

local function achievementOnEnter(cell)
	if module.db.profile.overTooltip and cell.cType1 == "achievement" then
		GameTooltip:SetOwner(module.tracker, module.trackerAnchor)
		GameTooltip:SetHyperlink(GetAchievementLink(cell.cValue1))
	end
end

local function achievementOnLeave(cell)
	GameTooltip:Hide()
end

local function buttonOnClick(button)
	local category = button.category
	if not category then return end
	if module.db.char.categoryDisplay[category] == 1 then
		module.db.char.categoryDisplay[category] = 0
	else
		module.db.char.categoryDisplay[category] = 1
	end
	module:updateAchievementTracker()
end

local function timedAchievementUpdate(self, elapsed)
	self.elapsed = self.elapsed + elapsed
	if self.elapsed >= 1 then
		local aid = self.cValue1
		module.achievements[aid].remainingTime = module.achievements[aid].remainingTime - self.elapsed
		if module.db.profile.colorizeATimer then
			self.fontString:SetText("|cff"..module.db.profile.defaultANameColor..module.achievements[aid].name.."|r |cff"..getColorFromCompletionHEX(module.achievements[aid].remainingTime / module.achievements[aid].duration)..SecondsToTime(module.achievements[aid].remainingTime).."|r")
		else
			self.fontString:SetText("|cff"..module.db.profile.defaultANameColor..module.achievements[aid].name.."|r |cff"..module.db.profile.defaultATimerColor..SecondsToTime(module.achievements[aid].remainingTime).."|r")
		end
		self.elapsed = 0
	end
end

local cProvider, cPrototype, basecCellPrototype = QTIP:CreateCellProvider(QTIP.LabelProvider)
function cPrototype:SetupCell(tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth, cType1, cType2, cValue1, cValue2)
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	elseif self.statusBar then
		core:recycleStatusBar(self.statusBar)
		self.statusBar = nil
	end
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font, lPad, rPad, maxWidth, minWidth)
	if not module.disableInteraction then
		self:EnableMouse(true)
		self.cType1, self.cValue1 = cType1, cValue1
		self:SetScript("OnEnter", achievementOnEnter)
		self:SetScript("OnLeave", achievementOnLeave)
		self:SetScript("OnMouseDown", achievementOnMouseDown)
		if cType2 == "icon" and cType1 == "category" then
			self.button = core:getCollapseButton()
			self.button:RegisterForClicks("AnyDown")
			self.button.iconTexture:SetTexture(cValue2)
			self.button:SetScript("OnClick", buttonOnClick)
			self.button.category = cValue1
			if module.db.profile.categoryCollapseButtonSize > height then
				self.button:SetHeight(height)
				self.button:SetWidth(height)
			else
				self.button:SetHeight(module.db.profile.categoryCollapseButtonSize)
				self.button:SetWidth(module.db.profile.categoryCollapseButtonSize)
			end
			self.button:SetParent(self)
			if module.category[cValue1] then
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT")
			else
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT", module.db.profile.subCategoryCellPadding, 0)
			end
			self.button:SetToplevel(true)
			self.button:Show()
		end
	else
		self:EnableMouse(false)
		self.cType1, self.cValue1 = nil, nil
		self:SetScript("OnEnter", nil)
		self:SetScript("OnLeave", nil)
		self:SetScript("OnMouseDown", nil)
	end
	if cType2 == "icon" and cType1 == "achievement" then
		if module.db.profile.displayAchievementImage then
			self.button = core:getCollapseButton()
			self.button.iconTexture:SetTexture(cValue2)
			if module.db.profile.achievementImageSize > height then
				self.button:SetHeight(height)
				self.button:SetWidth(height)
			else
				self.button:SetHeight(module.db.profile.achievementImageSize)
				self.button:SetWidth(module.db.profile.achievementImageSize)
			end
			self.button:SetParent(self)
			if module.achievements[cValue1].sc then
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT", module.subAchievementIconPadding, 0)
			else
				self.button:SetPoint("TOPLEFT", self, "TOPLEFT", module.achievementIconPadding, 0)
			end
			self.button:SetToplevel(true)
			self.button:Show()
		end
		if module.achievements[cValue1].remainingTime and module.achievements[cValue1].remainingTime > 0 then
			self:SetScript("OnUpdate", timedAchievementUpdate)
			self.elapsed = 0
		else
			self:SetScript("OnUpdate", nil)
		end
	elseif cType2 == "statusBar" then
		local sbHeight = module.db.profile.statusBarHeight + 2
		if sbHeight > height then
			height = sbHeight
			self:SetHeight(height)
		end
		self.statusBar = core:getStatusBar()
		self.statusBar:SetParent(self)
		self.statusBar:SetPoint("TOPLEFT", module.db.profile.leftStatusBarPadding, -1)
		self.statusBar:SetPoint("BOTTOMRIGHT", -module.db.profile.rightStatusBarPadding, 1)
		self.statusBar:SetMinMaxValues(0, module.achievements[cValue1].criteria[cValue2].reqQuantity)
		self.statusBar:SetValue(module.achievements[cValue1].criteria[cValue2].quantity)
		self.statusBar:SetStatusBarTexture(LSM:Fetch("statusbar", module.db.profile.statusBarTexture))
		self.statusBar.bg:SetTexture(LSM:Fetch("statusbar", module.db.profile.statusBarTexture))
		local r, g, b
		if module.db.profile.colorizeStatusBar then
			r, g, b = module.achievements[cValue1].criteria[cValue2].rColor, module.achievements[cValue1].criteria[cValue2].gColor, module.achievements[cValue1].criteria[cValue2].bColor
		else
			r, g, b = unpack(module.db.profile.defaultStatusBarColor)
		end
		self.statusBar.bg:SetVertexColor(r*0.33, g*0.33, b*0.33)
		self.statusBar:SetStatusBarColor(r, g, b)
		self.statusBar.status:SetFontObject(module.sbATrFontTQ2)
		if module.achievements[cValue1].criteria[cValue2].cType == 62 or module.achievements[cValue1].criteria[cValue2].cType == 67 then
			self.statusBar.status:SetText(core:formatMoney(module.achievements[cValue1].criteria[cValue2].quantity))
		elseif module.db.profile.colorizeStatusBarText then
			self.statusBar.status:SetText("|cff"..getColorFromCompletionHEX(module.achievements[cValue1].criteria[cValue2].quantity / module.achievements[cValue1].criteria[cValue2].reqQuantity)..module.achievements[cValue1].criteria[cValue2].quantity.." / "..module.achievements[cValue1].criteria[cValue2].reqQuantity.."|r")
		else
			self.statusBar.status:SetText("|cff"..module.db.profile.defaultStatusBarTextColor..module.achievements[cValue1].criteria[cValue2].quantity.." / "..module.achievements[cValue1].criteria[cValue2].reqQuantity.."|r")
		end
		self.statusBar:Show()
	end
	return width, height
end

function cPrototype:ReleaseCell()
	if self.button then
		core:recycleButton(self.button)
		self.button = nil
	elseif self.statusBar then
		core:recycleStatusBar(self.statusBar)
		self.statusBar = nil
	end
	self:EnableMouse(false)
	self.cType1, self.cValue1, self.elapsed = nil, nil, nil
	self:SetScript("OnEnter", nil)
	self:SetScript("OnLeave", nil)
	self:SetScript("OnMouseDown", nil)
	self:SetScript("OnUpdate", nil)
end

function module:addAchievementToTracker(aid, achievementPadding, criteriaPadding)
	local tempLine = self.tracker:AddLine()
	self.visibleAchievements[aid] = tempLine
	if self.achievements[aid].startTime then
		self.achievements[aid].remainingTime = self.achievements[aid].duration + self.achievements[aid].startTime - GetTime()
		if self.db.profile.colorizeATimer then
			self.tracker:SetCell(tempLine, 1, "|cff"..self.db.profile.defaultANameColor..self.achievements[aid].name.."|r |cff"..getColorFromCompletionHEX(self.achievements[aid].remainingTime / self.achievements[aid].duration)..SecondsToTime(self.achievements[aid].remainingTime).."|r", self.aATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns + 1, cProvider, achievementPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", "icon", aid, self.achievements[aid].image)
		else
			self.tracker:SetCell(tempLine, 1, "|cff"..self.db.profile.defaultANameColor..self.achievements[aid].name.."|r |cff"..self.db.profile.defaultATimerColor..SecondsToTime(self.achievements[aid].remainingTime).."|r", self.aATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns + 1, cProvider, achievementPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", "icon", aid, self.achievements[aid].image)
		end
	else
		self.tracker:SetCell(tempLine, 1, "|cff"..self.db.profile.defaultANameColor..self.achievements[aid].name.."|r", self.aATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns + 1, cProvider, achievementPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", "icon", aid, self.achievements[aid].image)
	end
	if self.achievements[aid].numCriteria == 0 then
		tempLine = self.tracker:AddLine()
		self.tracker:SetCell(tempLine, 1, self.achievements[aid].description, self.acATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns + 1, cProvider, criteriaPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", nil, aid)
	elseif self.achievements[aid].numCriteria == 1 then
		if self.achievements[aid].criteria[1].statusBar and self.db.profile.displayAsStatusBar then
			if self.db.profile.showSBDescription then
				tempLine = self.tracker:AddLine()
				self.tracker:SetCell(tempLine, 1, "|cff"..self.db.profile.defaultDescriptionColor..self.achievements[aid].description.."|r", self.acATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns + 1, cProvider, criteriaPadding, 0, self.totalMaxWidth, self.totalMinWidth, "achievement", nil, aid)
			end
			tempLine = self.tracker:AddLine()
			self.tracker:SetCell(tempLine, 1, " ", self.acATrFontTQ2, "CENTER", self.db.profile.numBooleanCriteriaColumns + 1, cProvider, 0, self.db.profile.rightCellPadding, self.totalMaxWidth, self.totalMinWidth, "achievement", "statusBar", aid, 1)
		else
			tempLine = self.tracker:AddLine()
			if self.db.profile.colorizeACrName then
				self.tracker:SetCell(tempLine, 1, "|cff"..self.achievements[aid].criteria[1].color..self.achievements[aid].description.."|r", self.acATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns, cProvider, criteriaPadding, 0, self.maxWidth, self.minWidth, "achievement", nil, aid)
			else
				self.tracker:SetCell(tempLine, 1, "|cff"..self.db.profile.defaultACrNameColor..self.achievements[aid].description.."|r", self.acATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns, cProvider, criteriaPadding, 0, self.maxWidth, self.minWidth, "achievement", nil, aid)
			end
			if self.db.profile.colorizeACrStatus then
				self.tracker:SetCell(tempLine, self.db.profile.numBooleanCriteriaColumns + 1, "|cff"..self.achievements[aid].criteria[1].color..self.achievements[aid].criteria[1].quantity.." / "..self.achievements[aid].criteria[1].reqQuantity.."|r", self.acATrFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "achievement", nil, aid)
			else
				self.tracker:SetCell(tempLine, self.db.profile.numBooleanCriteriaColumns + 1, "|cff"..self.db.profile.defaultACrStatusColor..self.achievements[aid].criteria[1].quantity.." / "..self.achievements[aid].criteria[1].reqQuantity.."|r", self.acATrFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "achievement", nil, aid)
			end
		end
		self.visibleCriteria[self.achievements[aid].criteria[1].id] = tempLine
	elseif self.achievements[aid].numCriteria > 1 then
		local cIndex = 1
		for cid, acData in pairs (self.achievements[aid].criteria) do
			if (self.db.profile.hideCompletedCriteria and acData.completed == 0) or (not self.db.profile.hideCompletedCriteria) then
				if acData.reqQuantity > 1 then
					if (acData.cType == 7 and acData.quantity > 0) or acData.cType ~= 7 then
						tempLine = self.tracker:AddLine()
						if self.db.profile.colorizeACrName then
							self.tracker:SetCell(tempLine, 1, "|cff"..acData.color..acData.name.."|r", self.acATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns, cProvider, criteriaPadding, 0, self.maxWidth, self.minWidth, "achievement", nil, aid)
						else
							self.tracker:SetCell(tempLine, 1, "|cff"..self.db.profile.defaultACrNameColor..acData.name.."|r", self.acATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns, cProvider, criteriaPadding, 0, self.maxWidth, self.minWidth, "achievement", nil, aid)
						end
						if self.db.profile.colorizeACrStatus then
							self.tracker:SetCell(tempLine, self.db.profile.numBooleanCriteriaColumns + 1, "|cff"..acData.color..acData.quantity.." / "..acData.reqQuantity.."|r", self.acATrFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "achievement", nil, aid)
						else
							self.tracker:SetCell(tempLine, self.db.profile.numBooleanCriteriaColumns + 1, "|cff"..self.db.profile.defaultACrStatusColor..acData.quantity.." / "..acData.reqQuantity.."|r", self.acATrFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "achievement", nil, aid)
						end
					end
					self.visibleCriteria[self.achievements[aid].criteria[cid].id] = tempLine.."~"..criteriaPadding
				else
					if cIndex == 1 then
						tempLine = self.tracker:AddLine()
						self.tracker:SetCell(tempLine, self.db.profile.numBooleanCriteriaColumns + 1, " ", self.acATrFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "achievement", nil, aid)
						if self.db.profile.colorizeACrName then
							self.tracker:SetCell(tempLine, cIndex, "|cff"..acData.color..acData.name.."|r", self.acATrFontTQ2, "LEFT", 1, cProvider, criteriaPadding, 0, self.columnMaxWidth, self.columnMinWidth, "achievement", nil, aid)
						else
							self.tracker:SetCell(tempLine, cIndex, "|cff"..self.db.profile.defaultACrNameColor..acData.name.."|r", self.acATrFontTQ2, "LEFT", 1, cProvider, criteriaPadding, 0, self.columnMaxWidth, self.columnMinWidth, "achievement", nil, aid)
						end
					else
						if self.db.profile.colorizeACrName then
							self.tracker:SetCell(tempLine, cIndex, "|cff"..acData.color..acData.name.."|r", self.acATrFontTQ2, "LEFT", 1, cProvider, 0, 0, self.columnMaxWidth, self.columnMinWidth, "achievement", nil, aid)
						else
							self.tracker:SetCell(tempLine, cIndex, "|cff"..self.db.profile.defaultACrNameColor..acData.name.."|r", self.acATrFontTQ2, "LEFT", 1, cProvider, 0, 0, self.columnMaxWidth, self.columnMinWidth, "achievement", nil, aid)
						end
					end
					self.visibleCriteria[self.achievements[aid].criteria[cid].id] = tempLine.."~"..criteriaPadding.."~"..cIndex
					if cIndex == self.db.profile.numBooleanCriteriaColumns then cIndex = 1 else cIndex = cIndex + 1 end
				end
			end
		end
	end
end

function module:updateAchievements()
	wipe(self.visibleAchievements)
	wipe(self.visibleCriteria)
	local tempLine
	if self.db.profile.showByCategory then
		for cid, cData in pairs(self.category) do
			self.db.char.categoryDisplay[cid] = self.db.char.categoryDisplay[cid] or 1
			tempLine = self.tracker:AddLine()
			if self.db.profile.colorizeACName then
				self.tracker:SetCell(tempLine, 1, "|cff"..cData.color..cData.name.."|r", self.cATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns, cProvider, self.categoryCellPadding, 0, self.maxWidth, self.minWidth, "category", "icon", cid, self.db.char.categoryDisplay[cid] == 0 and self.plus or self.minus)
			else
				self.tracker:SetCell(tempLine, 1, "|cff"..self.db.profile.defaultACrNameColor..cData.name.."|r", self.cATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns, cProvider, self.categoryCellPadding, 0, self.maxWidth, self.minWidth, "category", "icon", cid, self.db.char.categoryDisplay[cid] == 0 and self.plus or self.minus)
			end
			if self.db.profile.colorizeACCounter then
				self.tracker:SetCell(tempLine, self.db.profile.numBooleanCriteriaColumns + 1, "|cff"..cData.color..cData.numTCA.." / "..cData.numTA.."|r", self.cATrFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "category", nil, cid)
			else
				self.tracker:SetCell(tempLine, self.db.profile.numBooleanCriteriaColumns + 1, "|cff"..self.db.profile.defaultACCounterColor..cData.numTCA.." / "..cData.numTA.."|r", self.cATrFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "category", nil, cid)
			end
			if self.db.char.categoryDisplay[cid] == 1 or self.db.profile.disableInteraction then
				if cData.numAchievements > 0 then
					for aid in pairs(cData.achievements) do
						self:addAchievementToTracker(aid, self.achievementCellPadding, self.criteriaCellPadding)
					end
				end
				if cData.numSubCategory > 0 then
					for scid, scData in pairs(cData.subCategory) do
						self.db.char.categoryDisplay[scid] = self.db.char.categoryDisplay[scid] or 1
						tempLine = self.tracker:AddLine()
						if self.db.profile.colorizeASCName then
							self.tracker:SetCell(tempLine, 1, "|cff"..scData.color..scData.name.."|r", self.scATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns, cProvider, self.subCategoryCellPadding, 0, self.maxWidth, self.minWidth, "category", "icon", scid, self.db.char.categoryDisplay[scid] == 0 and self.plus or self.minus)
						else
							self.tracker:SetCell(tempLine, 1, "|cff"..self.db.profile.defaultASCNameColor..scData.name.."|r", self.scATrFontTQ2, "LEFT", self.db.profile.numBooleanCriteriaColumns, cProvider, self.subCategoryCellPadding, 0, self.maxWidth, self.minWidth, "category", "icon", scid, self.db.char.categoryDisplay[scid] == 0 and self.plus or self.minus)
						end
						if self.db.profile.colorizeASCCounter then
							self.tracker:SetCell(tempLine, self.db.profile.numBooleanCriteriaColumns + 1, "|cff"..scData.color..scData.numTCA.." / "..scData.numTA.."|r", self.scATrFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "category", nil, scid)
						else
							self.tracker:SetCell(tempLine, self.db.profile.numBooleanCriteriaColumns + 1, "|cff"..self.db.profile.defaultASCCounterColor..scData.numTCA.." / "..scData.numTA.."|r", self.scATrFontTQ2, "RIGHT", 1, cProvider, 0, self.db.profile.rightCellPadding, self.statusMaxWidth, self.statusMinWidth, "category", nil, scid)
						end
						if self.db.char.categoryDisplay[scid] == 1 or self.db.profile.disableInteraction then
							if scData.numAchievements > 0 then
								for aid in pairs(scData.achievements) do
									self:addAchievementToTracker(aid, self.subAchievementCellPadding, self.subCriteriaCellPadding)
								end
							end
						end
					end
				end
			end
		end
	else
		for aid, aData in pairs(self.achievements) do
			self:addAchievementToTracker(aid, self.achievementCellPadding, self.criteriaCellPadding)
		end
	end
end

function module:updateAchievementTracker()
	if self.db.profile.showInQuestTracker and core.moduleState["questsTracker"] then
		core.cbh:Fire("updateQuestsTracker")
	elseif self.db.profile.toggleTracker and self.numTrackedAchievement > 0 and not self.hideInCombat then
		self.tracker = QTIP:Acquire("tomQuest2Achievements", self.db.profile.numBooleanCriteriaColumns + 1)
		self.tracker:Clear()
		self.tracker:SetBackdropColor(self.db.profile.backDropColor[1], self.db.profile.backDropColor[2], self.db.profile.backDropColor[3], self.db.profile.backDropColor[4])
		self.tracker:SetBackdropBorderColor(self.db.profile.borderColor[1], self.db.profile.borderColor[2], self.db.profile.borderColor[3], self.db.profile.borderColor[4])
		self.tracker:SetScale(self.db.profile.scale)
		self.tracker:SetFrameStrata("BACKGROUND")
		self:updateAchievements()
		local tempLine = self.tracker:AddLine()
		for i = 1, self.db.profile.numBooleanCriteriaColumns do
			self.tracker:SetCell(tempLine, i, " ", self.fixATrFontTQ2, "LEFT", 1, nil, 0, 0, self.columnMaxWidth, self.columnMinWidth)
		end
		self.tracker:SetCell(tempLine, self.db.profile.numBooleanCriteriaColumns + 1, " ", self.fixATrFontTQ2, "RIGHT", 1, nil, 0, 0, self.tatusMaxWidth, self.statusMinWidth)
		self.tracker:ClearAllPoints()
		if self.db.profile.showAnchor then
			self.tracker:SetPoint(self.anchorSetPoint[self.db.profile.anchorPosition].point, self.anchor, self.anchorSetPoint[self.db.profile.anchorPosition].relPoint)
		else
			self.tracker:SetPoint(self.anchorSetPoint[self.db.profile.anchorPosition].point, self.anchor, self.anchorSetPoint[self.db.profile.anchorPosition].relPoint, self.anchorSetPoint[self.db.profile.anchorPosition].x, self.anchorSetPoint[self.db.profile.anchorPosition].y)
		end
		if not self.db.profile.disableInteraction then
			self.tracker:UpdateScrolling(self.db.profile.hasMaxHeight and self.db.profile.maxHeight or nil)
			self.tracker:EnableMouse(true)
		else
			self.tracker:EnableMouse(false)
		end
		self.tracker:Show()
	else
		QTIP:Release(self.tracker)
		self.tracker = nil
		return
	end
end

--------------------------
--- Keybinding globals ---
--------------------------
BINDING_NAME_TQ2ACHIEVEMENTTRACKERTOGGLE = core.LC["Achievement Tracker"]

function tQ2toggleAchievementTracker(display)
	if not display then
		module.db.profile.toggleTracker = not module.db.profile.toggleTracker
		if not module.db.profile.toggleTracker then
			QTIP:Release(module.tracker)
			module.tracker = nil
		else
			module:updateAchievementTracker()
		end
	elseif display == 1 and not module.db.profile.toggleTracker then
		module.db.profile.toggleTracker = not module.db.profile.toggleTracker
		module:updateAchievementTracker()
	elseif display == 0 and module.db.profile.toggleTracker then
		module.db.profile.toggleTracker = not module.db.profile.toggleTracker
		QTIP:Release(module.tracker)
		module.tracker = nil
	end
end